# Native imports
import ctypes
import ctypes.wintypes

# Third-party imports
import asyncio

# Custom imports
from wizwalker.windows.input import InputHandler

user32 = ctypes.windll.user32

keycode_map = {"TAB": 9, "ENTER": 13, "ALT": 18, "ESC": 27, "SPACE": 32}


class Keyboard(InputHandler):
    def __init__(self, window_handle):
        super().__init__(window_handle)
        self.window_handle = window_handle

    def type_key(self, char):
        code = None

        try:
            code = keycode_map[char]
        except KeyError:
            code = ord(char)

        user32.PostMessageW(self.window_handle, 0x102, code, 0)

    def type_keycode(self, code):
        user32.PostMessageW(self.window_handle, 0x102, code, 0)

    def type_string(self, string):
        for s in string:
            self.type_key(s)


if __name__ == "__main__":
    """ Some tests """
    from wizwalker.utils import get_all_wizard_handles

    try:
        window_handle = get_all_wizard_handles()[0]
    except IndexError:
        print("No running wizard windows")
        exit(0)

    async def say_hello():

        keyboard = Keyboard(window_handle)

        await keyboard.send_key("ENTER", 0.01)
        keyboard.type_string("Hello world!")
        await keyboard.send_key("ENTER", 0.1)

    asyncio.run(say_hello())



# Native imports
import ctypes
import os, sys

# Third-party imports
import cv2
import numpy
import asyncio
import wizwalker
from wizwalker.utils import calculate_perfect_yaw
from simple_chalk import chalk

# Custom imports
from .utils import get_all_wiz_handles, XYZYaw
from .pixels import DeviceContext, match_image
from .keyboard import Keyboard
from .mouse import Mouse
from .window import Window
from .battle import Battle
from .card import Card


SPELLS_FOLDER = "spells"


user32 = ctypes.windll.user32
__DIRNAME__ = os.path.dirname(__file__)
all_clients = []


async def unregister_all():
    print("Un-hooking all registered clients")
    for client in all_clients:
        await client.unregister()


class Client(DeviceContext, Keyboard, Window):
    def __init__(self, handle=None):
        # Inherit all methods from Parent Classes
        super().__init__(handle)
        self.window_handle = handle
        self.logging = True
        self.name = None
        # rectangles defined as (x, y, width, height)
        self._friends_area = (625, 65, 20, 240)
        self._spell_area = (245, 290, 370, 70)
        self._confirm_area = (355, 370, 100, 70)

        self.walker = None
        self.mouse = Mouse(handle)

    @classmethod
    def register(cls, nth=0, name=None, handle=None, hooks=[]):
        """
        Assigns the instance to a wizard101 window (Required before using any other API functions)
        """
        client = cls()
        global all_clients

        if handle:
            client.window_handle = handle
        else:
            handles = get_all_wiz_handles()

            if len(handles) == 0:
                client.log("No Wizard101 windows detected")
                return None

            handles.sort()
            # Assigns the one at index nth
            client.window_handle = handles[nth]

        # A window exists, add it to global variable
        all_clients.append(client)

        # Start `_anti_disconnect` task
        client._anti_disconnect_task = asyncio.create_task(client._anti_disconnect())

        client.walker = wizwalker.Client(client.window_handle)

        client.activate_hooks = client.walker.activate_hooks

        client.mouse = Mouse(client.window_handle)

        if name:
            client.set_name(name)

        return client

    def set_name(self, name):
        self.name = name
        user32.SetWindowTextW(self.window_handle, f"[{name}] Wizard101")

    def log(self, message):
        if self.logging:
            s = ""
            if self.name != None:
                s += chalk.magentaBright(f"[{self.name}] ")
            s += message
            print(s)
            sys.stdout.flush()

    async def _anti_disconnect(self):
        while True:
            # Wait 10 minute
            await asyncio.sleep(10 * 60)
            # Sent o key to stay awake
            await self.send_key("O", 0.1)

    async def unregister(self):
        """ Properly unregister hooks and more """
        user32.SetWindowTextW(self.window_handle, "Wizard101")
        self._anti_disconnect_task.cancel()
        await self.walker.close()
        return 1

    async def wait(self, s):
        """ Alias for asyncio.sleep() that return self for function chaining """
        await asyncio.sleep(s)
        return self

    """
    STATE DETECTION
    """

    def is_crown_shop(self):
        """ Matches a red pixel in the close icon of the opened crown shop menu """
        return self.pixel_matches_color((788, 53), (197, 40, 41), 10)

    def is_idle(self):
        """ Matches pixels in the spell book (only visible when not in battle) """
        spellbook_yellow = self.pixel_matches_color(
            (781, 551), (255, 251, 64), tolerance=20
        )
        spellbook_brown = self.pixel_matches_color(
            (728, 587), (79, 29, 29), tolerance=20
        )
        return spellbook_brown and spellbook_yellow

    def is_dialog_more(self):
        more_lower_right = self.pixel_matches_color((669, 611), (112, 32, 54), 15)
        more_top_left = self.pixel_matches_color((585, 609), (115, 32, 56), 15)
        return more_lower_right and more_top_left

    def is_health_low(self):
        # Matches a pixel in the lower third of the health globe
        POSITION = (23, 563)
        COLOR = (126, 41, 3)
        TOLERANCE = 15
        return not self.pixel_matches_color(POSITION, COLOR, tolerance=TOLERANCE)

    def is_mana_low(self):
        # Matches a pixel in the lower third of the mana globe
        POSITION = (79, 591)
        COLOR = (66, 13, 83)
        TOLERANCE = 15
        return not self.pixel_matches_color(POSITION, COLOR, tolerance=TOLERANCE)

    def is_press_x(self):
        x_area = self.get_image(region=(350, 540, 100, 20))
        found = match_image(x_area, __DIRNAME__ + "/images/x.png")
        return found != False

    def get_confirm(self):
        confirm_img = self.get_image(self._confirm_area)
        found = match_image(
            confirm_img, __DIRNAME__ + "/images/confirm.png", threshold=0.2
        )
        if found:
            x = found[0] + self._confirm_area[0]
            y = found[1] + self._confirm_area[1]
            found = (x, y)

        return found

    """
    ACTIONS BASED ON STATES
    """

    async def use_potion_if_needed(self):
        mana_low = self.is_mana_low()
        health_low = self.is_health_low()

        if mana_low:
            self.log("Mana is low, using potion")
        if health_low:
            self.log("Health is low, using potion")
        if mana_low or health_low:
            await self.mouse.click(160, 590, delay=0.2)

    async def finish_loading(self):
        self.log("Awaiting loading")
        while self.is_idle():
            await asyncio.sleep(0.2)

        while not self.is_idle() or self.is_crown_shop():
            await asyncio.sleep(0.5)

        await asyncio.sleep(1)

    async def go_through_dialog(self, times=1):
        # Wait for press X, or more/done button
        self.log("Going through dialog")
        while times >= 1:
            times -= 1
            while (not self.is_press_x()) and (not self.is_dialog_more()):
                await asyncio.sleep(0.5)

            if self.is_press_x():
                await self.send_key("X", 0.1)

            while not self.is_dialog_more():
                await asyncio.sleep(0.5)

            while self.is_dialog_more():
                await self.send_key("SPACEBAR", 0.1)
                await asyncio.sleep(0.1)

        await asyncio.sleep(1)

    async def logout_and_in(self, confirm=False):
        self.log("Loging out")
        await self.send_key("ESC", 0.1)
        await self.mouse.click(259, 506, delay=0.3)
        if confirm:
            await self.click_confirm()
        # wait for player select screen
        print("Wait for loading")
        while not (self.pixel_matches_color((361, 599), (133, 36, 62), tolerance=20)):
            await self.wait(0.5)

        print("Log back in")
        await self.mouse.click(395, 594)
        await self.finish_loading()
        if self.is_crown_shop():
            await self.wait(0.5)
            await self.send_key("ESC", 0.1)
            await self.send_key("ESC", 0.1)

    async def press_x(self):
        while not self.is_press_x():
            await self.wait(0.5)
        await self.send_key("X", 0.1)

    async def click_confirm(self):
        await self.wait(0.2)  #
        confirm = self.get_confirm()
        while not confirm:
            await self.wait(0.5)
            confirm = self.get_confirm()

        await self.mouse.click(*confirm, duration=0.2, delay=0.2)
        await self.wait(0.5)

    """
    POSITION & MOVEMENT
    """

    async def get_quest_xyz(self):
        """
        Returns the X, Y, Z coordinates to the quest destionation
        Requires the `quest_struct` hook to be activated
        """
        return await self.walker.quest_xyz()

    async def teleport_to(self, location: XYZYaw):
        """
        Teleports to XYZYaw location
        Will return immediately if player movement is locked
        Requires the `player_struct` hook to be activated
        """
        if await self.walker.move_lock():
            return

        await self.walker.teleport(**location._asdict())
        await self.send_key("W", 0.1)

    async def walk_to(self, location):
        """
        Walks to XYZYaw location in a **straight** line only.
        Will _not_ work if there are obstacles in the way. Ideal for short distances
        Will return immediately if player movement is locked
        Requires the `player_struct` hook to be activated
        """
        if await self.walker.move_lock():
            return

        await self.walker.goto(location.x, location.y)

    async def teleport_to_friend(self, match_img):
        """
        Completes a set of actions to teleport to a friend.
        The friend must have the proper symbol next to it
        symbol must match the image passed as 'match_img'

        """
        self.set_active()
        # Check if friends already opened (and close it)
        while self.pixel_matches_color((780, 364), (230, 0, 0), 40):
            await self.mouse.click(780, 364, duration=0.1)
            await self.wait(0.2)

        # Open friend menu
        await self.mouse.click(780, 50, duration=0.2)
        await self.wait(0.2)

        # Find friend that matches friend match_img
        friend_area_img = self.get_image(region=self._friends_area)

        found = match_image(friend_area_img, match_img, threshold=0.2)

        if found is not False:
            _, y = found
            offset_y = self._friends_area[1]

            # Select friend
            await self.mouse.click(670, offset_y + y, duration=0.2, delay=0.5)
            # Select port
            await self.mouse.click(450, 115, duration=0.2, delay=0.5)
            # Select yes
            await self.click_confirm()
            await self.wait(1)

            return self
        else:
            self.log("Friend cound not be found")
            return False

    async def face_quest_destination(self):
        """
        Changes the player's yaw to be facing the quest destination.
        *Note:* depending on your location, this may differ from where your quest arrow is pointing to
        """
        xyz = await self.walker.xyz()
        quest_xyz = await self.walker.quest_xyz()
        yaw = calculate_perfect_yaw(xyz, quest_xyz)
        await self.walker.set_yaw(yaw)

    """
    BATTLE ACTIONS & METHODS
    """

    def get_battle(self, name=None) -> Battle:
        """
        Returns a `Battle` object linked to this client
        """
        return Battle(self, name)

    async def find_spell(self, spell_name, threshold=0.1) -> Card:
        """
        Searches spell area for an image matching `spell_name`
        Returns x positions of spell if found
        returns None if not found
        """
        # Move mouse out of area to get a clear image
        await self.mouse.move_out(self._spell_area)

        # Get screenshot of `spell_area`
        b_spell_area = self.get_image(self._spell_area)

        extensions = [".png", ".jpg", "jpeg", ".bmp"]
        file_name = spell_name
        if not spell_name[-4:] in extensions:
            file_name += ".png"

        spell_path = os.path.join(SPELLS_FOLDER, file_name)

        res = match_image(b_spell_area, spell_path, threshold, debug=False)

        if res:
            x, y = res
            # We're only interested in the x position
            offset_x = self._spell_area[0]
            # a card width is 52 pixels, round to the nearest 1/2 card (26 pixels)
            adjusted_x = round(x / 26) * 26

            spell_pos = offset_x + adjusted_x
            return Card(self, spell_name, spell_pos)
        else:
            return None

    async def pass_turn(self) -> None:
        """
        Clicks `pass` while in a battle
        """
        await self.mouse.click(254, 398, duration=0.2, delay=0.5)
        await self.wait(0.5)


def register_clients(
    n_handles_expected: int, names: list = [], confirm_position: bool = False
):
    """
    n_handles_expected: the expected # of wiz windows opened. Use -1 for undetermined
    names: A list of strings that will serve as the names of the windows
    """
    accepted = False
    while not accepted:
        handles = get_all_wiz_handles()
        n_handles = len(handles)

        if n_handles != n_handles_expected and n_handles_expected > 0:
            print(
                f"Invalid number of windows open. {n_handles_expected} required, {n_handles} detected."
            )
            os.system("pause")
            exit()
        else:
            print(f"{n_handles} windows detected")

        # Fill names array if necessary
        for i in range(n_handles - len(names)):
            names.append(None)

        # Register and order the windows from left to right, top to bottom
        w = [Client.register(handle=handles[i]) for i in range(n_handles)]

        # Sort
        def sortFunc(win):
            rect = win.get_rect()
            round_y = (rect[1] // 100) * 100
            return rect[0] + (round_y * 10)

        w.sort(key=sortFunc)

        # Set names
        for i in range(len(w)):
            w[i].set_name(names[i])

        # Confirm position
        if confirm_position:
            print("Is this order ok?")
            answer = input("[y] or n: ")
            if answer.lower().strip()[0] == "y" or answer == "":
                accepted = True
            else:
                print("Re-order the windows")
                os.system("pause")
        else:
            accepted = True

    # Returns the sorted clients in an array
    return w




# Native imports
import ctypes
import ctypes.wintypes

# Third-party imports
import numpy as np
import cv2

# Custom imports
from .window import Window

user32 = ctypes.WinDLL("user32.dll")
gdi32 = ctypes.WinDLL("gdi32.dll")

# Converted to python from https://docs.microsoft.com/en-us/windows/win32/gdi/capturing-an-image
# by Starrfox


class BITMAPINFOHEADER(ctypes.Structure):
    _fields_ = [
        ("biSize", ctypes.c_uint32),
        ("biWidth", ctypes.c_int),
        ("biHeight", ctypes.c_int),
        ("biPlanes", ctypes.c_short),
        ("biBitCount", ctypes.c_short),
        ("biCompression", ctypes.c_uint32),
        ("biSizeImage", ctypes.c_uint32),
        ("biXPelsPerMeter", ctypes.c_long),
        ("biYPelsPerMeter", ctypes.c_long),
        ("biClrUsed", ctypes.c_uint32),
        ("biClrImportant", ctypes.c_uint32),
    ]


class BITMAP(ctypes.Structure):
    _fields_ = [
        ("bmType", ctypes.c_long),
        ("bmWidth", ctypes.c_long),
        ("bmHeight", ctypes.c_long),
        ("bmWidthBytes", ctypes.c_long),
        ("bmPlanes", ctypes.wintypes.WORD),
        ("bmBitsPixel", ctypes.wintypes.WORD),
        ("bmBits", ctypes.wintypes.LPVOID),
    ]


class DeviceContext(Window):
    def __init__(self, handle):
        super().__init__(handle)
        self.window_handle = handle

    def get_image(self, region=None):
        _, _, w, h = self.get_rect()
        x, y = 0, 0

        if region and len(region) == 4:
            x, y, w, h = region

        # Get devices context
        wDC = user32.GetWindowDC(self.window_handle)
        # Where we will move the pixels to
        mDC = gdi32.CreateCompatibleDC(wDC)

        gdi32.SetStretchBltMode(wDC, 4)
        # Create empty bitmap
        mBM = gdi32.CreateCompatibleBitmap(wDC, w, h)

        # Select wDC into bitmap
        gdi32.SelectObject(mDC, mBM)

        gdi32.BitBlt(mDC, 0, 0, w, h, wDC, x, y, 0x00CC0020)

        bitmap = BITMAP()

        # Don't know what that does
        gdi32.GetObjectA(mBM, ctypes.sizeof(BITMAP), ctypes.byref(bitmap))

        bi = BITMAPINFOHEADER()
        bi.biSize = ctypes.sizeof(BITMAPINFOHEADER)
        bi.biWidth = bitmap.bmWidth
        bi.biHeight = bitmap.bmHeight
        bi.biPlanes = 1
        bi.biBitCount = 32
        bi.biCompression = 0
        bi.biSizeImage = 0
        bi.biXPelsPerMeter = 0
        bi.biYPelsPerMeter = 0
        bi.biClrUsed = 0
        bi.biClrImportant = 0

        bitmap_size = (
            (
                (bitmap.bmWidth * bitmap.bmBitsPixel + bitmap.bmBitsPixel - 1)
                // bitmap.bmBitsPixel
            )
            * 4
            * bitmap.bmHeight
        )

        bitmap_buffer = (ctypes.c_char * bitmap_size)()

        gdi32.GetDIBits(
            wDC,
            mBM,
            0,
            bitmap.bmHeight,
            ctypes.byref(bitmap_buffer),
            ctypes.byref(bi),
            0,
        )

        img = np.frombuffer(bitmap_buffer.raw, dtype="uint8")
        img = img.reshape((bitmap.bmHeight, bitmap.bmWidth, 4))
        img = np.flip(img, 0)
        # Remove the alpha channel
        img = img[:, :, :3]
        return img

    def get_pixel(self, x, y):
        hDC = user32.GetWindowDC(self.window_handle)
        rgb = gdi32.GetPixel(hDC, x, y)
        user32.ReleaseDC(self.window_handle, hDC)
        r = rgb & 0xFF
        g = (rgb >> 8) & 0xFF
        b = (rgb >> 16) & 0xFF
        return (r, g, b)

    def screenshot(self, name, region=None):
        image = self.get_image(region=region)
        cv2.imshow("mat", image)
        cv2.waitKey(0)
        cv2.imwrite(name, image)

    def pixel_matches_color(self, xy, expected_rgb, tolerance=0):
        pixel = self.get_pixel(*xy)
        if len(pixel) == 3 or len(expected_rgb) == 3:  # RGB mode
            r, g, b = pixel[:3]
            exR, exG, exB = expected_rgb[:3]
            return (
                (abs(r - exR) <= tolerance)
                and (abs(g - exG) <= tolerance)
                and (abs(b - exB) <= tolerance)
            )
        else:
            assert (
                False
            ), f"Color mode was expected to be length 3 (RGB), but pixel is length {len(pix)} and expected_RGB is length { len(expected_rgb)}"


def to_cv2_img(data):
    if type(data) is str:
        # It's a file name
        img = cv2.imread(data)
        if img is None:
            print(f"Unable to find `{data}`")
        return img

    elif type(data) is np.ndarray:
        # It's a np array
        return data

    return None


def match_image(largeImg, smallImg, threshold=0.1, debug=False):
    """ 
    Finds smallImg in largeImg using template matching 
    Adjust threshold for the precision of the match (between 0 and 1, the lowest being more precise)
    Returns false if no match was found with the given threshold
    """

    method = cv2.TM_SQDIFF_NORMED

    small_image = to_cv2_img(smallImg)
    large_image = to_cv2_img(largeImg)

    if (small_image is None) or (large_image is None):
        print("Error: large_image or small_image is None")
        return False

    w, h = small_image.shape[:-1]

    result = cv2.matchTemplate(small_image, large_image, method)

    # We want the minimum squared difference
    mn, _, mnLoc, _ = cv2.minMaxLoc(result)

    if mn >= threshold:
        if debug:
            cv2.imshow("output", large_image)
            cv2.waitKey(0)
        return False

    # Extract the coordinates of our best match
    x, y = mnLoc

    if debug:
        # Draw the rectangle:
        # Get the size of the template. This is the same size as the match.
        trows, tcols = small_image.shape[:2]

        # If I don't call this a get a TypeError :P
        large_image = np.array(large_image)
        # Draw the rectangle on large_image
        cv2.rectangle(large_image, (x, y), (x + tcols, y + trows), (0, 0, 255), 2)

        # Display the original image with the rectangle around the match.
        cv2.imshow("output", large_image)

        # The image is only displayed if we call this
        cv2.waitKey(0)

    # Return coordinates to center of match
    return (x + (w * 0.5), y + (h * 0.5))




# Native imports
import ctypes
from collections import namedtuple

# Third-party imports
import asyncio


XYZYaw = namedtuple("XYZYaw", "x y z yaw")

user32 = ctypes.windll.user32


def get_all_wiz_handles():
    """
    Retrieves all window handles for windows that have the 
    'Wizard Graphical Client' class
    """
    target_class = "Wizard Graphical Client"

    handles = []

    # callback takes a window handle and an lparam and returns true/false on if we should keep going
    # iterating
    # https://docs.microsoft.com/en-us/previous-versions/windows/desktop/legacy/ms633498(v=vs.85)
    def callback(handle, _):
        class_name = ctypes.create_unicode_buffer(len(target_class))
        user32.GetClassNameW(handle, class_name, len(target_class) + 1)
        if target_class == class_name.value:
            handles.append(handle)

        # iterate all windows
        return 1

    # https://docs.python.org/3/library/ctypes.html#callback-functions
    enumwindows_func_type = ctypes.WINFUNCTYPE(
        ctypes.c_bool,  # return type
        ctypes.c_int,  # arg1 type
        ctypes.POINTER(ctypes.c_int),  # arg2 type
    )

    # Transform callback into a form we can pass to the dll
    callback = enumwindows_func_type(callback)

    # EnumWindows takes a callback every iteration is passed to
    # and an lparam
    # https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-enumwindows
    user32.EnumWindows(callback, 0)

    return handles


def count_wiz_clients():
    return len(get_all_wiz_handles())


async def finish_all_loading(*players):
    await asyncio.gather(*[player.finish_loading() for player in players])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wizsdk']

package_data = \
{'': ['*'], 'wizsdk': ['images/*']}

install_requires = \
['asyncchain>=0.1.0,<0.2.0',
 'asyncio>=3.4.3,<4.0.0',
 'numpy>=1.19.1,<2.0.0',
 'opencv-python>=4.4.0,<5.0.0',
 'simple_chalk>=0.1.0,<0.2.0',
 'wizwalker>=0.13.0,<0.14.0']

setup_kwargs = {
    'name': 'wizsdk',
    'version': '0.4.4',
    'description': 'API for interacting with and making bots for wizard101',
    'long_description': '# WizAPI botting tool\n\n### Create powerful bots for Wizard101\n\nA wrapper for [WizWalker](https://pypi.org/project/wizwalker/) to create the bots _you_ want, right now.\n',
    'author': 'Underpaid Dev',
    'author_email': 'underpaiddev1@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# Native imports
import sys

# Third-party imports
import asyncio
from simple_chalk import chalk

# Custom imports
from .pixels import DeviceContext, match_image
from .card import Card

# TODO
# Find enemy
# Find friend


class Battle(DeviceContext):
    def __init__(self, client, name=None):
        super().__init__(client.window_handle)
        self.round_count = 1
        self.client = client
        self.logging = self.client.logging
        self.is_idle = self.client.is_idle
        self.name = name

        self.is_over = False
        self.in_progress = False

        # rectangles defined as (x, y, width, height)
        self._spell_area = (245, 290, 370, 70)
        self._enemy_area = (68, 26, 650, 50)
        self._ally_area = (140, 580, 650, 50)

    async def loop(self):
        """
        Takes care of the looping logic for a battle
        """
        if not self.in_progress and not self.is_over:
            await self.start()
            return True
        else:
            await self.next_turn()
            return not self.is_over

    def log(self, message):
        if self.logging:
            s = ""
            if self.name != None:
                s += chalk.blueBright(f"[{self.name}] ")
            s += message
            print(s)
            sys.stdout.flush()

    def print_round(self):
        self.log(f"----------- Battle round {self.round_count} -----------")

    def is_turn(self) -> bool:
        """
        Returns if it's our turn to play
        by matching pixels in the `flee` button
        """
        flee_yellow = self.pixel_matches_color((550, 400), (255, 255, 0), tolerance=10)
        flee_brown = self.pixel_matches_color((578, 394), (124, 68, 0), tolerance=20)
        return flee_yellow and flee_brown

    async def start(self) -> None:
        """
        Waits for the first round then signals to the class that the battle has started.
        used in the `loop()` method
        """
        while not self.is_turn():
            await asyncio.sleep(1)

        self.is_over = False
        self.in_progress = True
        self.log("Battle is starting")
        self.print_round()

    async def next_turn(self) -> None:
        """
        Sleeps until next round or fight has ended
        Sets `is_over` to True if the fight has ended
        Increase the `round_count` otherwise
        Used in the `loop()` method
        """
        while self.is_turn():
            await asyncio.sleep(1)

        while not self.is_turn() and not self.is_idle():
            await asyncio.sleep(1)

        if self.is_idle():
            self.log("Battle has finished")
            self.is_over = True
            self.in_progress = False
        else:
            self.round_count += 1
            self.print_round()

    def get_enemy_positions(self):
        """
        Returns a list of length 4 with 0, if no enemy is present at the 
        location, or 1 if an enemy is present
        """
        Y = 75
        COLOR = (207, 186, 135)
        enemies = []
        for i in range(4):
            X = (174 * (i)) + 203
            enemy_present = 0
            if self.pixel_matches_color((X, Y), COLOR, tolerance=30):
                enemy_present = 1
            enemies.append(enemy_present)

        return enemies

    def get_enemy_count(self):
        """
        Returns the number enemies in the fight
        """
        return sum(self.get_enemy_positions())

    def find_enemy(self, enemy_image):
        """ 
        Attemps to find the position of an enemy that matches the image provided 
        returns 0, 1, 2, 3 if found otherwise returns False
        """
        enemy_area = self.get_image(region=self._enemy_area)

        found = match_image(enemy_area, enemy_image, threshold=0.2)

        if found:
            return round((found[0] - 60) / 170)

        return False

    def find_ally(self, ally_image):
        """ 
        Attemps to find the position of an ally the matches the image provided 
        returns 4, 5, 6, 7 if found otherwise returns False
        """
        ally_area = self.get_image(region=self._ally_area)

        found = match_image(ally_area, ally_image, threshold=0.2)

        if found:
            return 7 - round((found[0] - 100) / 170)

        return False
